@@ -38,7 +38,7 @@
 #include <openssl/err.h>
 #include <openssl/engine.h>
 
-#include "openconnect.h"
+#include "openconnect-internal.h"
 
 static int proxy_write(int fd, unsigned char *buf, size_t len);
 
@@ -132,8 +132,6 @@ static int process_http_response(struct openconnect_info *vpninfo, int *result,
 	while ((i = openconnect_SSL_gets(vpninfo->https_ssl, buf, sizeof(buf)))) {
 		char *colon;
 
-		vpninfo->progress(vpninfo, PRG_TRACE, "%s\n", buf);
-
 		if (i < 0) {
 			vpninfo->progress(vpninfo, PRG_ERR, "Error processing HTTP response\n");
 			return -EINVAL;
@@ -147,6 +145,38 @@ static int process_http_response(struct openconnect_info *vpninfo, int *result,
 		if (*colon == ' ')
 			colon++;
 
+		/* Handle Set-Cookie first so that we can avoid printing the
+		   webvpn cookie in the verbose debug output */
+		if (!strcasecmp(buf, "Set-Cookie")) {
+			char *semicolon = strchr(colon, ';');
+			char *print_equals, *equals = strchr(colon, '=');
+			int ret;
+
+			if (semicolon)
+				*semicolon = 0;
+
+			if (!equals) {
+				vpninfo->progress(vpninfo, PRG_ERR, "Invalid cookie offered: %s\n", buf);
+				return -EINVAL;
+			}
+			*(equals++) = 0;
+
+			print_equals = equals;
+			/* Don't print the webvpn cookie unless it's empty; we don't
+			   want people posting it in public with debugging output */
+			if (!strcmp(colon, "webvpn") && *equals)
+				print_equals = "<elided>";
+			vpninfo->progress(vpninfo, PRG_TRACE, "%s: %s=%s%s%s\n",
+					  buf, colon, print_equals, semicolon?";":"",
+					  semicolon?(semicolon+1):"");
+
+			ret = http_add_cookie(vpninfo, colon, equals);
+			if (ret)
+				return ret;
+		} else {
+			vpninfo->progress(vpninfo, PRG_TRACE, "%s: %s\n", buf, colon);
+		}
+
 		if (!strcasecmp(buf, "Connection")) {
 			if (!strcasecmp(colon, "Close"))
 				closeconn = 1;
@@ -173,24 +203,6 @@ static int process_http_response(struct openconnect_info *vpninfo, int *result,
 				return -EINVAL;
 			}
 		}
-		if (!strcasecmp(buf, "Set-Cookie")) {
-			char *semicolon = strchr(colon, ';');
-			char *equals = strchr(colon, '=');
-			int ret;
-
-			if (semicolon)
-				*semicolon = 0;
-
-			if (!equals) {
-				vpninfo->progress(vpninfo, PRG_ERR, "Invalid cookie offered: %s\n", buf);
-				return -EINVAL;
-			}
-			*(equals++) = 0;
-
-			ret = http_add_cookie(vpninfo, colon, equals);
-			if (ret)
-				return ret;
-		}
 		if (!strcasecmp(buf, "Transfer-Encoding")) {
 			if (!strcasecmp(colon, "chunked"))
 				bodylen = BODY_CHUNKED;
@@ -366,7 +378,7 @@ static int run_csd_script(struct openconnect_info *vpninfo, char *buf, int bufle
 	char fname[16];
 	int fd, ret;
 
-	if (!vpninfo->uid_csd_given) {
+	if (!vpninfo->uid_csd_given && !vpninfo->csd_wrapper) {
 		vpninfo->progress(vpninfo, PRG_ERR,
 				  "Error: Server asked us to download and run a 'Cisco Secure Desktop' trojan.\n"
 				  "This facility is disabled by default for security reasons, so you may wish to enable it.");
@@ -424,7 +436,7 @@ static int run_csd_script(struct openconnect_info *vpninfo, char *buf, int bufle
 				exit(1);
 			}
 		}
-		if (vpninfo->uid_csd == 0) {
+		if (vpninfo->uid_csd == 0 && !vpninfo->csd_wrapper) {
 			fprintf(stderr, "Warning: you are running insecure "
 				"CSD code with root privileges\n"
 				"\t Use command line option \"--csd-user\"\n");
@@ -434,6 +446,8 @@ static int run_csd_script(struct openconnect_info *vpninfo, char *buf, int bufle
 			   on stdout, which the CSD trojan spews. */
 			dup2(2, 1);
 		}
+		if (vpninfo->csd_wrapper)
+			csd_argv[i++] = vpninfo->csd_wrapper;
 		csd_argv[i++] = fname;
 		csd_argv[i++] = "-ticket";
 		if (asprintf(&csd_argv[i++], "\"%s\"", vpninfo->csd_ticket) == -1)
@@ -468,8 +482,8 @@ static int run_csd_script(struct openconnect_info *vpninfo, char *buf, int bufle
 		csd_argv[i++] = "-langselen";
 		csd_argv[i++] = NULL;
 
-		execv(fname, csd_argv);
-		vpninfo->progress(vpninfo, PRG_ERR, "Failed to exec CSD script %s\n", fname);
+		execv(csd_argv[0], csd_argv);
+		vpninfo->progress(vpninfo, PRG_ERR, "Failed to exec CSD script %s\n", csd_argv[0]);
 		exit(1);
 	}
 
@@ -506,8 +520,8 @@ char *local_strcasestr(const char *haystack, const char *needle)
 #define strcasestr local_strcasestr
 #endif
 
-int parse_url(char *url, char **res_proto, char **res_host, int *res_port,
-	      char **res_path, int default_port)
+int internal_parse_url(char *url, char **res_proto, char **res_host,
+		       int *res_port, char **res_path, int default_port)
 {
 	char *proto = url;
 	char *host, *path, *port_str;
@@ -585,6 +599,10 @@ int openconnect_obtain_cookie(struct openconnect_info *vpninfo)
 	char *method = "GET";
 
  retry:
+	if (form_buf) {
+		free(form_buf);
+		form_buf = NULL;
+	}
 	if (!vpninfo->https_ssl && openconnect_open_https(vpninfo)) {
 		vpninfo->progress(vpninfo, PRG_ERR, "Failed to open HTTPS connection to %s\n",
 			vpninfo->hostname);
@@ -650,7 +668,7 @@ int openconnect_obtain_cookie(struct openconnect_info *vpninfo)
 			free(vpninfo->urlpath);
 			vpninfo->urlpath = NULL;
 
-			ret = parse_url(vpninfo->redirect_url, NULL, &host, &port, &vpninfo->urlpath, 0);
+			ret = internal_parse_url(vpninfo->redirect_url, NULL, &host, &port, &vpninfo->urlpath, 0);
 			if (ret) {
 				vpninfo->progress(vpninfo, PRG_ERR, "Failed to parse redirected URL '%s': %s\n",
 						  vpninfo->redirect_url, strerror(-ret));
@@ -697,7 +715,9 @@ int openconnect_obtain_cookie(struct openconnect_info *vpninfo)
 			vpninfo->redirect_url = NULL;
 			goto retry;
 		} else {
-			char *lastslash = strrchr(vpninfo->urlpath, '/');
+			char *lastslash = NULL;
+			if (vpninfo->urlpath)
+				lastslash = strrchr(vpninfo->urlpath, '/');
 			if (!lastslash) {
 				free(vpninfo->urlpath);
 				vpninfo->urlpath = vpninfo->redirect_url;
@@ -1055,9 +1075,9 @@ int process_proxy(struct openconnect_info *vpninfo, int ssl_sock)
 	return -EIO;
 }
 
-int set_http_proxy(struct openconnect_info *vpninfo, char *proxy)
+int openconnect_set_http_proxy(struct openconnect_info *vpninfo, char *proxy)
 {
-	char *url = strdup(proxy);
+	char *url = proxy;
 	int ret;
 
 	if (!url)
@@ -1068,8 +1088,8 @@ int set_http_proxy(struct openconnect_info *vpninfo, char *proxy)
 	free(vpninfo->proxy);
 	vpninfo->proxy = NULL;
 
-	ret = parse_url(url, &vpninfo->proxy_type, &vpninfo->proxy,
-			&vpninfo->proxy_port, NULL, 80);
+	ret = internal_parse_url(url, &vpninfo->proxy_type, &vpninfo->proxy,
+				 &vpninfo->proxy_port, NULL, 80);
 	if (ret)
 		goto out;
 
