@@ -36,7 +36,6 @@
 #include <sys/syslog.h>
 #include <sys/utsname.h>
 #include <sys/types.h>
-#include <openssl/rand.h>
 #include <openssl/ui.h>
 #ifdef OPENCONNECT_LIBPROXY
 #include LIBPROXY_HDR
@@ -45,7 +44,7 @@
 #define _GNU_SOURCE
 #include <getopt.h>
 
-#include "openconnect.h"
+#include "openconnect-internal.h"
 
 static int write_new_config(struct openconnect_info *vpninfo, char *buf, int buflen);
 static void write_progress(struct openconnect_info *info, int level, const char *fmt, ...);
@@ -57,6 +56,30 @@ int background;
 int do_passphrase_from_fsid;
 int nocertcheck;
 
+enum {
+	OPT_AUTHGROUP = 0x100,
+	OPT_CAFILE,
+	OPT_COOKIEONLY,
+	OPT_COOKIE_ON_STDIN,
+	OPT_CSD_USER,
+	OPT_CSD_WRAPPER,
+	OPT_DISABLE_IPV6,
+	OPT_DTLS_CIPHERS,
+	OPT_FORCE_DPD,
+	OPT_KEY_PASSWORD_FROM_FSID,
+	OPT_LIBPROXY,
+	OPT_NO_CERT_CHECK,
+	OPT_NO_DTLS,
+	OPT_NO_HTTP_KEEPALIVE,
+	OPT_NO_PASSWD,
+	OPT_NO_PROXY,
+	OPT_PASSWORD_ON_STDIN,
+	OPT_PRINTCOOKIE,
+	OPT_RECONNECT_TIMEOUT,
+	OPT_SERVERCERT,
+	OPT_USERAGENT,
+};
+
 static struct option long_options[] = {
 	{"background", 0, 0, 'b'},
 	{"certificate", 1, 0, 'c'},
@@ -78,28 +101,30 @@ static struct option long_options[] = {
 	{"user", 1, 0, 'u'},
 	{"verbose", 0, 0, 'v'},
 	{"version", 0, 0, 'V'},
-	{"cafile", 1, 0, '0'},
-	{"no-dtls", 0, 0, '1'},
-	{"cookieonly", 0, 0, '2'},
-	{"printcookie", 0, 0, '3'},
+	{"cafile", 1, 0, OPT_CAFILE},
+	{"no-dtls", 0, 0, OPT_NO_DTLS},
+	{"cookieonly", 0, 0, OPT_COOKIEONLY},
+	{"printcookie", 0, 0, OPT_PRINTCOOKIE},
 	{"quiet", 0, 0, 'q'},
 	{"queue-len", 1, 0, 'Q'},
 	{"xmlconfig", 1, 0, 'x'},
-	{"cookie-on-stdin", 0, 0, '4'},
-	{"passwd-on-stdin", 0, 0, '5'},
-	{"no-passwd", 0, 0, '6'},
-	{"reconnect-timeout", 1, 0, '7'},
-	{"dtls-ciphers", 1, 0, '8'},
-	{"authgroup", 1, 0, '9'},
-	{"servercert", 1, 0, 0x01},
-	{"key-password-from-fsid", 0, 0, 0x02},
-	{"useragent", 1, 0, 0x03},
-	{"csd-user", 1, 0, 0x04},
-	{"disable-ipv6", 0, 0, 0x05},
-	{"no-proxy", 0, 0, 0x06},
-	{"libproxy", 0, 0, 0x07},
-	{"no-http-keepalive", 0, 0, 0x08},
-	{"no-cert-check", 0, 0, 0x09},
+	{"cookie-on-stdin", 0, 0, OPT_COOKIE_ON_STDIN},
+	{"passwd-on-stdin", 0, 0, OPT_PASSWORD_ON_STDIN},
+	{"no-passwd", 0, 0, OPT_NO_PASSWD},
+	{"reconnect-timeout", 1, 0, OPT_RECONNECT_TIMEOUT},
+	{"dtls-ciphers", 1, 0, OPT_DTLS_CIPHERS},
+	{"authgroup", 1, 0, OPT_AUTHGROUP},
+	{"servercert", 1, 0, OPT_SERVERCERT},
+	{"key-password-from-fsid", 0, 0, OPT_KEY_PASSWORD_FROM_FSID},
+	{"useragent", 1, 0, OPT_USERAGENT},
+	{"csd-user", 1, 0, OPT_CSD_USER},
+	{"csd-wrapper", 1, 0, OPT_CSD_WRAPPER},
+	{"disable-ipv6", 0, 0, OPT_DISABLE_IPV6},
+	{"no-proxy", 0, 0, OPT_NO_PROXY},
+	{"libproxy", 0, 0, OPT_LIBPROXY},
+	{"no-http-keepalive", 0, 0, OPT_NO_HTTP_KEEPALIVE},
+	{"no-cert-check", 0, 0, OPT_NO_CERT_CHECK},
+	{"force-dpd", 1, 0, OPT_FORCE_DPD},
 	{NULL, 0, 0, 0},
 };
 
@@ -115,12 +140,14 @@ void usage(void)
 	printf("      --cookie-on-stdin           Read cookie from standard input\n");
 	printf("  -d, --deflate                   Enable compression (default)\n");
 	printf("  -D, --no-deflate                Disable compression\n");
+	printf("      --force-dpd=INTERVAL        Set minimum Dead Peer Detection interval\n");
 	printf("  -g, --usergroup=GROUP           Set login usergroup\n");
 	printf("  -h, --help                      Display help text\n");
 	printf("  -i, --interface=IFNAME          Use IFNAME for tunnel interface\n");
 	printf("  -l, --syslog                    Use syslog for progress messages\n");
 	printf("  -U, --setuid=USER               Drop privileges after connecting\n");
 	printf("      --csd-user=USER             Drop privileges during CSD execution\n");
+	printf("      --csd-wrapper=SCRIPT        Run SCRIPT instead of CSD binary\n");
 	printf("  -m, --mtu=MTU                   Request MTU from server\n");
 	printf("  -p, --key-password=PASS         Set key passphrase or TPM SRK PIN\n");
 	printf("      --key-password-from-fsid    Key passphrase is fsid of file system\n");
@@ -150,7 +177,7 @@ void usage(void)
 	printf("      --no-cert-check             Do not require server SSL cert to be valid\n");
 	printf("      --passwd-on-stdin           Read password from standard input\n");
 	printf("      --reconnect-timeout         Connection retry timeout in seconds\n");
-	printf("      --servercert                Server's certificate SHA1 fingerprint\n");
+	printf("      --servercert=FINGERPRINT    Server's certificate SHA1 fingerprint\n");
 	printf("      --useragent=STRING          HTTP header User-Agent: field\n");
 	exit(1);
 }
@@ -215,10 +242,6 @@ int main(int argc, char **argv)
 	vpninfo->uid_csd_given = 0;
 	vpninfo->validate_peer_cert = validate_peer_cert;
 
-	if (RAND_bytes(vpninfo->dtls_secret, sizeof(vpninfo->dtls_secret)) != 1) {
-		fprintf(stderr, "Failed to initialise DTLS secret\n");
-		exit(1);
-	}
 	if (!uname(&utsbuf))
 		vpninfo->localname = utsbuf.nodename;
 	else
@@ -230,41 +253,41 @@ int main(int argc, char **argv)
 			break;
 
 		switch (opt) {
-		case '0':
+		case OPT_CAFILE:
 			vpninfo->cafile = optarg;
 			break;
-		case 0x01:
+		case OPT_SERVERCERT:
 			vpninfo->servercert = optarg;
 			break;
-		case '1':
+		case OPT_NO_DTLS:
 			vpninfo->dtls_attempt_period = 0;
 			break;
-		case '2':
+		case OPT_COOKIEONLY:
 			cookieonly = 1;
 			break;
-		case '3':
+		case OPT_PRINTCOOKIE:
 			cookieonly = 2;
 			break;
-		case '4':
+		case OPT_COOKIE_ON_STDIN:
 			read_stdin(&vpninfo->cookie);
 			/* If the cookie is empty, ignore it */
 			if (! *vpninfo->cookie) {
 				vpninfo->cookie = NULL;
 			}
 			break;
-		case '5':
+		case OPT_PASSWORD_ON_STDIN:
 			read_stdin(&vpninfo->password);
 			break;
-		case '6':
+		case OPT_NO_PASSWD:
 			vpninfo->nopasswd = 1;
 			break;
-		case '7':
+		case OPT_RECONNECT_TIMEOUT:
 			vpninfo->reconnect_timeout = atoi(optarg);
 			break;
-		case '8':
+		case OPT_DTLS_CIPHERS:
 			vpninfo->dtls_ciphers = optarg;
 			break;
-		case '9':
+		case OPT_AUTHGROUP:
 			vpninfo->authgroup = optarg;
 			break;
 		case 'b':
@@ -324,10 +347,10 @@ int main(int argc, char **argv)
 			proxy = optarg;
 			autoproxy = 0;
 			break;
-		case 0x06:
+		case OPT_NO_PROXY:
 			autoproxy = 0;
 			proxy = NULL;
-		case 0x07:
+		case OPT_LIBPROXY:
 #ifndef OPENCONNECT_LIBPROXY
 			fprintf(stderr, "This version of openconnect was built without libproxy support\n");
 			exit(1);
@@ -335,12 +358,12 @@ int main(int argc, char **argv)
 			autoproxy = 1;
 			proxy = NULL;
 			break;
-		case 0x08:
+		case OPT_NO_HTTP_KEEPALIVE:
 			fprintf(stderr, "Disabling all HTTP connection re-use due to --no-http-keepalive option.\n"
 				"If this helps, please report to <openconnect-devel@lists.infradead.org>.\n");
 			vpninfo->no_http_keepalive = 1;
 			break;
-		case 0x09:
+		case OPT_NO_CERT_CHECK:
 			nocertcheck = 1;
 			break;
 		case 's':
@@ -366,7 +389,7 @@ int main(int argc, char **argv)
 			}
 			break;
 		}
-		case 0x04: {
+		case OPT_CSD_USER: {
 			char *strend;
 			vpninfo->uid_csd = strtol(optarg, &strend, 0);
 			if (strend[0]) {
@@ -381,7 +404,10 @@ int main(int argc, char **argv)
 			vpninfo->uid_csd_given = 1;
 			break;
 		}
-		case 0x05:
+		case OPT_CSD_WRAPPER:
+			vpninfo->csd_wrapper = optarg;
+			break;
+		case OPT_DISABLE_IPV6:
 			vpninfo->disable_ipv6 = 1;
 			break;
 		case 'Q':
@@ -404,13 +430,16 @@ int main(int argc, char **argv)
 			vpninfo->xmlconfig = optarg;
 			vpninfo->write_new_config = write_new_config;
 			break;
-		case 0x02:
+		case OPT_KEY_PASSWORD_FROM_FSID:
 			do_passphrase_from_fsid = 1;
 			break;
-		case 0x03:
+		case OPT_USERAGENT:
 			free(vpninfo->useragent);
 			vpninfo->useragent = optarg;
 			break;
+		case OPT_FORCE_DPD:
+			vpninfo->dtls_times.dpd = vpninfo->ssl_times.dpd = atoi(optarg);
+			break;
 		default:
 			usage();
 		}
@@ -430,7 +459,7 @@ int main(int argc, char **argv)
 	if (autoproxy)
 		vpninfo->proxy_factory = px_proxy_factory_new();
 #endif
-	if (proxy && set_http_proxy(vpninfo, proxy))
+	if (proxy && openconnect_set_http_proxy(vpninfo, strdup(proxy)))
 		exit(1);
 
 	if (use_syslog) {
@@ -445,7 +474,7 @@ int main(int argc, char **argv)
 	sigaction(SIGUSR2, &sa, NULL);
 
 	if (vpninfo->sslkey && do_passphrase_from_fsid)
-		passphrase_from_fsid(vpninfo);
+		openconnect_passphrase_from_fsid(vpninfo);
 
 	if (config_lookup_host(vpninfo, argv[optind]))
 		exit(1);
@@ -455,7 +484,7 @@ int main(int argc, char **argv)
 		char *scheme;
 		char *group;
 
-		if (parse_url(url, &scheme, &vpninfo->hostname, &vpninfo->port,
+		if (internal_parse_url(url, &scheme, &vpninfo->hostname, &vpninfo->port,
 			      &group, 443)) {
 			fprintf(stderr, "Failed to parse server URL '%s'\n",
 				url);
@@ -601,7 +630,7 @@ static int validate_peer_cert(struct openconnect_info *vpninfo, X509 *peer_cert,
 	if (nocertcheck)
 		return 0;
 
-	ret = get_cert_sha1_fingerprint(vpninfo, peer_cert, fingerprint);
+	ret = openconnect_get_cert_sha1(vpninfo, peer_cert, fingerprint);
 	if (ret)
 		return ret;
 
