@@ -49,7 +49,7 @@
 #include <openssl/pkcs12.h>
 #include <openssl/x509v3.h>
 
-#include "openconnect.h"
+#include "openconnect-internal.h"
 
 /* OSX < 1.6 doesn't have AI_NUMERICSERV */
 #ifndef AI_NUMERICSERV
@@ -173,8 +173,8 @@ static int load_pkcs12_certificate(struct openconnect_info *vpninfo, PKCS12 *p12
 		return -EINVAL;
 	}
 	if (cert) {
+		vpninfo->cert_x509 = cert;
 		SSL_CTX_use_certificate(vpninfo->https_ctx, cert);
-		X509_free(cert);
 	} else {
 		vpninfo->progress(vpninfo, PRG_ERR,
 				  "PKCS#12 contained no certificate!");
@@ -265,6 +265,28 @@ static int load_tpm_certificate(struct openconnect_info *vpninfo)
 	return 0;
 }
 
+static int reload_pem_cert(struct openconnect_info *vpninfo)
+{
+	BIO *b = BIO_new(BIO_s_file_internal());
+
+	if (!b)
+		return -ENOMEM;
+
+	if (BIO_read_filename(b, vpninfo->cert) <= 0) {
+	err:
+		BIO_free(b);
+		vpninfo->progress(vpninfo, PRG_ERR,
+				 "Failed to reload X509 cert for expiry check\n");
+		report_ssl_errors(vpninfo);
+		return -EIO;
+	}
+	vpninfo->cert_x509 = PEM_read_bio_X509_AUX(b, NULL, NULL, NULL);
+	if (!vpninfo->cert_x509)
+		goto err;
+
+	return 0;
+}
+
 static int load_certificate(struct openconnect_info *vpninfo)
 {
 	vpninfo->progress(vpninfo, PRG_TRACE,
@@ -306,6 +328,9 @@ static int load_certificate(struct openconnect_info *vpninfo)
 		return -EINVAL;
 	}
 
+	/* Ew, we can't get it back from the OpenSSL CTX in any sane fashion */
+	reload_pem_cert(vpninfo);
+
 	if (vpninfo->cert_type == CERT_TYPE_UNKNOWN) {
 		FILE *f = fopen(vpninfo->sslkey, "r");
 		char buf[256];
@@ -372,48 +397,32 @@ static int load_certificate(struct openconnect_info *vpninfo)
 	return 0;
 }
 
-enum cert_hash_type {
-	EVP_MD5,
-	EVP_SHA1
-};
-
 static int get_cert_fingerprint(struct openconnect_info *vpninfo,
-				X509 *cert, enum cert_hash_type hash,
+				X509 *cert, const EVP_MD *type,
 				char *buf)
 {
 	unsigned char md[EVP_MAX_MD_SIZE];
 	unsigned int i, n;
 
-	switch (hash) {
-	case EVP_MD5:
-		if (!X509_digest(cert, EVP_md5(), md, &n))
-			return -ENOMEM;
-		break;
-	case EVP_SHA1:
-		if (!X509_digest(cert, EVP_sha1(), md, &n))
-			return -ENOMEM;
-		break;
-	default:
-		vpninfo->progress(vpninfo, PRG_ERR,
-				  "Unsupported SSL certificate hash function type\n");
-	}
+	if (!X509_digest(cert, type, md, &n))
+		return -ENOMEM;
 
-	for (i=0; i < n; i++) {
+	for (i=0; i < n; i++)
 		sprintf(&buf[i*2], "%02X", md[i]);
-	}
+
 	return 0;
 }
 
 int get_cert_md5_fingerprint(struct openconnect_info *vpninfo,
 			     X509 *cert, char *buf)
 {
-	return get_cert_fingerprint(vpninfo, cert, EVP_MD5, buf);
+	return get_cert_fingerprint(vpninfo, cert, EVP_md5(), buf);
 }
 
-int get_cert_sha1_fingerprint(struct openconnect_info *vpninfo,
+int openconnect_get_cert_sha1(struct openconnect_info *vpninfo,
 			      X509 *cert, char *buf)
 {
-	return get_cert_fingerprint(vpninfo, cert, EVP_SHA1, buf);
+	return get_cert_fingerprint(vpninfo, cert, EVP_sha1(), buf);
 }
 
 static int check_server_cert(struct openconnect_info *vpninfo, X509 *cert)
@@ -421,7 +430,7 @@ static int check_server_cert(struct openconnect_info *vpninfo, X509 *cert)
 	char fingerprint[EVP_MAX_MD_SIZE * 2 + 1];
 	int ret;
 
-	ret = get_cert_sha1_fingerprint(vpninfo, cert, fingerprint);
+	ret = openconnect_get_cert_sha1(vpninfo, cert, fingerprint);
 	if (ret)
 		return ret;
 
@@ -596,7 +605,7 @@ int match_cert_hostname(struct openconnect_info *vpninfo, X509 *peer_cert)
 			if (strlen(str) != len)
 				continue;
 
-			if (parse_url(str, &url_proto, &url_host, &url_port, &url_path, 0)) {
+			if (internal_parse_url(str, &url_proto, &url_host, &url_port, &url_path, 0)) {
 				OPENSSL_free(str);
 				continue;
 			}
@@ -781,6 +790,50 @@ static int ssl_app_verify_callback(X509_STORE_CTX *ctx, void *arg)
 }
 #endif
 
+static int check_certificate_expiry(struct openconnect_info *vpninfo)
+{
+	ASN1_TIME *notAfter;
+	char *reason = NULL;
+	time_t t;
+	int i;
+	if (!vpninfo->cert_x509)
+		return 0;
+
+	t = time(NULL);
+	notAfter = X509_get_notAfter(vpninfo->cert_x509);
+	i = X509_cmp_time(notAfter, &t);
+	if (!i) {
+		vpninfo->progress(vpninfo, PRG_ERR, "Error in client cert notAfter field\n");
+		return -EINVAL;
+	} else if (i < 0) {
+		reason = "has expired";
+	} else {
+		t += 60 * 86400;
+		i = X509_cmp_time(notAfter, &t);
+		if (i < 0) {
+			reason = "expires soon";
+		}
+	}
+	if (reason) {
+		BIO *bp = BIO_new(BIO_s_mem());
+		BUF_MEM *bm;
+		char *expiry = "<error>";
+		char zero = 0;
+
+		if (bp) {
+			ASN1_TIME_print(bp, notAfter);
+			BIO_write(bp, &zero, 1);
+			BIO_get_mem_ptr(bp, &bm);
+			expiry = bm->data;
+		}
+		vpninfo->progress(vpninfo, PRG_ERR, "Client certificate %s at: %s\n",
+				  reason, expiry);
+		if (bp)
+			BIO_free(bp);
+	}
+	return 0;
+}
+
 int openconnect_open_https(struct openconnect_info *vpninfo)
 {
 	method_const SSL_METHOD *ssl3_method;
@@ -850,7 +903,7 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 				    (!strncmp(proxies[i], "http://", 7) ||
 				     !strncmp(proxies[i], "socks://", 8) ||
 				     !strncmp(proxies[i], "socks5://", 9)))
-					parse_url(proxies[i], &vpninfo->proxy_type,
+					internal_parse_url(proxies[i], &vpninfo->proxy_type,
 						  &vpninfo->proxy, &vpninfo->proxy_port,
 						  NULL, 0);
 				i++;
@@ -943,7 +996,7 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 		}
 	}
 
-	ssl3_method = TLSv1_client_method();
+	ssl3_method = SSLv3_client_method();
 	if (!vpninfo->https_ctx) {
 		vpninfo->https_ctx = SSL_CTX_new(ssl3_method);
 
@@ -954,6 +1007,7 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 						  "Loading certificate failed. Aborting.\n");
 				return err;
 			}
+			check_certificate_expiry(vpninfo);
 		}
 
 		/* We just want to do:
@@ -970,8 +1024,15 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 #endif
 		SSL_CTX_set_default_verify_paths(vpninfo->https_ctx);
 
-		if (vpninfo->cafile)
-			SSL_CTX_load_verify_locations(vpninfo->https_ctx, vpninfo->cafile, NULL);
+		if (vpninfo->cafile) {
+			if (!SSL_CTX_load_verify_locations(vpninfo->https_ctx, vpninfo->cafile, NULL)) {
+				vpninfo->progress(vpninfo, PRG_ERR, "Failed to open CA file '%s'\n",
+						  vpninfo->cafile);
+				report_ssl_errors(vpninfo);
+				close(ssl_sock);
+				return -EINVAL;
+			}
+		}
 
 	}
 	https_ssl = SSL_new(vpninfo->https_ctx);
@@ -1026,7 +1087,7 @@ void openconnect_init_openssl(void)
 }
 
 #if defined(__sun__) || defined(__NetBSD__) || defined(__DragonFly__)
-int passphrase_from_fsid(struct openconnect_info *vpninfo)
+int openconnect_passphrase_from_fsid(struct openconnect_info *vpninfo)
 {
 	struct statvfs buf;
 
@@ -1040,7 +1101,7 @@ int passphrase_from_fsid(struct openconnect_info *vpninfo)
 	return 0;
 }
 #else
-int passphrase_from_fsid(struct openconnect_info *vpninfo)
+int openconnect_passphrase_from_fsid(struct openconnect_info *vpninfo)
 {
 	struct statfs buf;
 	unsigned *fsid = (unsigned *)&buf.f_fsid;
