@@ -47,7 +47,7 @@
 #endif
 #endif
 
-#include "openconnect.h"
+#include "openconnect-internal.h"
 
 /*
  * If an if_tun.h include file was found anywhere (by the Makefile), it's 
@@ -407,12 +407,25 @@ int setup_tun(struct openconnect_info *vpninfo)
 	} else {
 #ifdef IFF_TUN /* Linux */
 		struct ifreq ifr;
+		int tunerr;
 
 		tun_fd = open("/dev/net/tun", O_RDWR);
 		if (tun_fd < 0) {
+			/* Android has /dev/tun instead of /dev/net/tun
+			   Since other systems might have too, just try it
+			   as a fallback instead of using ifdef __ANDROID__ */
+			tunerr = errno;
+			tun_fd = open("/dev/tun", O_RDWR);
+		}
+		if (tun_fd < 0) {
+			/* If the error on /dev/tun is ENOENT, that's boring.
+			   Use the error we got on /dev/net/tun instead */
+			if (errno != -ENOENT)
+				tunerr = errno;
+
 			vpninfo->progress(vpninfo, PRG_ERR,
 					  "Failed to open tun device: %s\n",
-					  strerror(errno));
+					  strerror(tunerr));
 			exit(1);
 		}
 		memset(&ifr, 0, sizeof(ifr));
@@ -559,8 +572,10 @@ int tun_mainloop(struct openconnect_info *vpninfo, int *timeout)
 		while ((len = read(vpninfo->tun_fd, buf, sizeof(buf))) > 0) {
 			unsigned char *pkt = buf;
 #ifdef TUN_HAS_AF_PREFIX
-			pkt += 4;
-			len -= 4;
+			if (!vpninfo->script_tun) {
+				pkt += 4;
+				len -= 4;
+			}
 #endif
 			if (queue_new_packet(&vpninfo->outgoing_queue, pkt,
 					     len))
@@ -586,27 +601,29 @@ int tun_mainloop(struct openconnect_info *vpninfo, int *timeout)
 		int len = this->len;
 
 #ifdef TUN_HAS_AF_PREFIX
-		struct ip *iph = (void *)data;
-		int type;
+		if (!vpninfo->script_tun) {
+			struct ip *iph = (void *)data;
+			int type;
 
-		if (iph->ip_v == 6)
-			type = AF_INET6;
-		else if (iph->ip_v == 4)
-			type = AF_INET;
-		else {
-			static int complained = 0;
-			if (!complained) {
-				complained = 1;
-				vpninfo->progress(vpninfo, PRG_ERR,
-						  "Unknown packet (len %d) received: %02x %02x %02x %02x...\n",
-						  len, data[0], data[1], data[2], data[3]);
+			if (iph->ip_v == 6)
+				type = AF_INET6;
+			else if (iph->ip_v == 4)
+				type = AF_INET;
+			else {
+				static int complained = 0;
+				if (!complained) {
+					complained = 1;
+					vpninfo->progress(vpninfo, PRG_ERR,
+							  "Unknown packet (len %d) received: %02x %02x %02x %02x...\n",
+							  len, data[0], data[1], data[2], data[3]);
+				}
+				free(this);
+				continue;
 			}
-			free(this);
-			continue;
+			data -= 4;
+			len += 4;
+			*(int *)data = htonl(type);
 		}
-		data -= 4;
-		len += 4;
-		*(int *)data = htonl(type);
 #endif
 		vpninfo->incoming_queue = this->next;
 
